@@ -61,7 +61,6 @@
 #include "mount.h"
 #include "util.h"
 #include "resolve_host.h"
-#include "data_blob.h"
 
 #ifndef MS_MOVE 
 #define MS_MOVE 8192 
@@ -270,7 +269,7 @@ static int mount_usage(FILE * stream)
 	fprintf(stream,
 		"\n\tmapchars,nomapchars,nolock,servernetbiosname=<SRV_RFC1001NAME>");
 	fprintf(stream,
-		"\n\tdirectio,nounix,cifsacl,sec=<authentication mechanism>,sign,fsc");
+		"\n\tdirectio,nounix,cifsacl,sec=<authentication mechanism>,sign,seal,fsc");
 	fprintf(stream,
 		"\n\nOptions not needed for servers supporting CIFS Unix extensions");
 	fprintf(stream,
@@ -1003,6 +1002,7 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 				"WARNING: CIFS mount option 'fmask' is\
 				 deprecated. Use 'file_mode' instead.\n");
 			data = "file_mode";	/* BB fix this */
+			/* Fallthrough */
 		case OPT_FILE_MODE:
 			if (!value || !*value) {
 				fprintf(stderr,
@@ -1023,6 +1023,7 @@ parse_options(const char *data, struct parsed_mount_info *parsed_info)
 				"WARNING: CIFS mount option 'dmask' is\
 				 deprecated. Use 'dir_mode' instead.\n");
 			data = "dir_mode";
+			/* Fallthrough */
 		case OPT_DIR_MODE:
 			if (!value || !*value) {
 				fprintf(stderr,
